/* CAUTION: This is the ANSI C (only) version of the Numerical Recipes
   utility file complex.c.  Do not confuse this file with the same-named
   file complex.c that is supplied in the same subdirectory or archive
   as the header file complex.h.  *That* file contains both ANSI and
   traditional K&R versions, along with #ifdef macros to select the
   correct version.  *This* file contains only ANSI C.               */

#include <math.h>
#include "nr.h"


dcomplex dCadd(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r+b.r;
	c.i=a.i+b.i;
	return c;
}

dcomplex dCsub(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r-b.r;
	c.i=a.i-b.i;
	return c;
}


dcomplex dCmul(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r*b.r-a.i*b.i;
	c.i=a.i*b.r+a.r*b.i;
	return c;
}

dcomplex dComplex(double re, double im)
{
	dcomplex c;
	c.r=re;
	c.i=im;
	return c;
}

dcomplex dConjg(dcomplex z)
{
	dcomplex c;
	c.r=z.r;
	c.i = -z.i;
	return c;
}

dcomplex dCdiv(dcomplex a, dcomplex b)
{
	dcomplex c;
	double r,den;
	if (fabs(b.r) >= fabs(b.i)) {
		r=b.i/b.r;
		den=b.r+r*b.i;
		c.r=(a.r+r*a.i)/den;
		c.i=(a.i-r*a.r)/den;
	} else {
		r=b.r/b.i;
		den=b.i+r*b.r;
		c.r=(a.r*r+a.i)/den;
		c.i=(a.i*r-a.r)/den;
	}
	return c;
}

double dCabs(dcomplex z)
{
	double x,y,ans,temp;
	x=fabs(z.r);
	y=fabs(z.i);
	if (x == 0.0)
		ans=y;
	else if (y == 0.0)
		ans=x;
	else if (x > y) {
		temp=y/x;
		ans=x*sqrt(1.0+temp*temp);
	} else {
		temp=x/y;
		ans=y*sqrt(1.0+temp*temp);
	}
	return ans;
}

dcomplex dCsqrt(dcomplex z)
{
	dcomplex c;
	double x,y,w,r;
	if ((z.r == 0.0) && (z.i == 0.0)) {
		c.r=0.0;
		c.i=0.0;
		return c;
	} else {
		x=fabs(z.r);
		y=fabs(z.i);
		if (x >= y) {
			r=y/x;
			w=sqrt(x)*sqrt(0.5*(1.0+sqrt(1.0+r*r)));
		} else {
			r=x/y;
			w=sqrt(y)*sqrt(0.5*(r+sqrt(1.0+r*r)));
		}
		if (z.r >= 0.0) {
			c.r=w;
			c.i=z.i/(2.0*w);
		} else {
			c.i=(z.i >= 0.0) ? w : -w;
			c.r=z.i/(2.0*c.i);
		}
		return c;
	}
}

dcomplex dRCmul(double x, dcomplex a)
{
	dcomplex c;
	c.r=x*a.r;
	c.i=x*a.i;
	return c;
}
/* (C) Copr. 1986-92 Numerical Recipes Software s%{,*;15M. */

dcomplex dICmul( double y, dcomplex a )
{
  dcomplex c;
  c.r = -a.i*y;
  c.i = a.r*y;
  return c;
}
/* (C) 1995 Maureen O'Leary and David Boas, The Moon */


/* additional functions: Copyright by Martin Ostermeyer 1995 */

double dCphas(dcomplex z)
{
  return atan2(z.i,z.r); 
}

dcomplex dCpolar(double abs, double phase)
{
  dcomplex c;

  c.r = (double)( ((double)abs)*cos((double)phase) );
  c.i = (double)( ((double)abs)*sin((double)phase) );
  return c;
}


double dsmall_phase_difference(double a, double b)
{
  double d1,d2;

  if(a>0.0){
    d1 = a - 360.0 - b;
    d2 = a - b;
    if(fabs(d1) < fabs(d2))
      return d1;
    else
      return d2;
  }
  else {
    d1 = a + 360.0 - b;
    d2 = a - b;
    if(fabs(d1) < fabs(d2))
      return d1;
    else
      return d2;
  }
}




/*****************************************************
  SPHERICAL HARMONICS
*****************************************************/
dcomplex dSphHarmonic( int l, double theta )
{
  double p, p1, p2, x;
  int i;

  x = cos( theta );
  p2 = 1;
  p1 = x;

  if( l==0 ) return dComplex(sqrt(1.0/(4.0*3.14159))*p2,0.0);
  if( l==1 ) return dComplex(sqrt(3.0/(4.0*3.14159))*p1,0.0);
  for( i=1; i<l; i++ )
    {
      p = ((2*i+1)*x*p1 - i*p2)/(i+1);
      p2 = p1;
      p1 = p;
    }
  return dComplex(sqrt((2*l+1)/(4.0*3.14159))*p,0.0);
}

/*****************************************************
  COMPLEX EXPONENTIAL
*****************************************************/
dcomplex dCexp( dcomplex x )
{
  dcomplex z;
  z.r = cos(x.i) * exp(x.r);
  z.i = sin(x.i) * exp(x.r);
  return z;
}

/*****************************************************
  COMPLEX COSINE
*****************************************************/
dcomplex dCcos( dcomplex x )
{
  dcomplex z;
  z.r = ( cos(x.r) * exp(-x.i) + cos(x.r) * exp(x.i) ) / 2.0;
  z.i = ( sin(x.r) * exp(-x.i) - sin(x.r) * exp(x.i) ) / 2.0;
  return z;
}

/*****************************************************
  COMPLEX SINE
*****************************************************/
dcomplex dCsin( dcomplex x )
{
  dcomplex z;
  z.r = ( sin(x.r) * exp(-x.i) + sin(x.r) * exp(x.i) ) / 2.0;
  z.i = -1 * ( cos(x.r) * exp(-x.i) - cos(x.r) * exp(x.i) ) / 2.0;
  return z;
}

/*****************************************************
  COMPLEX LOG
*****************************************************/
dcomplex dClog( dcomplex x )
{
  dcomplex z;
  z = dComplex( log(dCabs(x)), atan2(x.i,x.r) );
  return z;
}
